import io
from typing import Optional, Union

from pyrogram import Client
from pyrogram.types import Message


async def send_as_file(
    self,
    chat_id: Union[int, str],
    text: str,
    filename: str = "output.txt",
    caption: str = "",
    log: Union[bool, str] = False,
    reply_to_message_id: Optional[int] = None,
) -> "Message":
    """\nYou can send large outputs as file
    Example:
            @userge.send_as_file(chat_id=12345, text="hello")
    Parameters:
        chat_id (``int`` | ``str``):
            Unique identifier (int) or username (str) of the target chat.
            For your personal cloud (Saved Messages)
            you can simply use "me" or "self".
            For a contact that exists in your Telegram address book
            you can use his phone number (str).
        text (``str``):
            Text of the message to be sent.
        filename (``str``, *optional*):
            file_name for output file.
        caption (``str``, *optional*):
            caption for output file.
        log (``bool`` | ``str``, *optional*):
            If ``True``, the message will be forwarded
            to the log channel.
            If ``str``, the logger name will be updated.
        reply_to_message_id (``int``, *optional*):
            If the message is a reply, ID of the original message.
    Returns:
        On success, the sent Message is returned.
    """
    doc = io.BytesIO(text.encode())
    doc.name = filename

    return await self.send_document(
        chat_id=chat_id,
        document=doc,
        caption=caption[:1024],
        disable_notification=True,
        reply_to_message_id=reply_to_message_id,
    )


Client.send_as_file = send_as_file



import asyncio
from typing import Union

from pyrogram import Client
from pyrogram.types import Message


async def edit_message_text(
    self,
    chat_id: Union[int, str],
    message_id: int,
    text: str,
    del_in: int = 0,
    *args,
    **kwargs
) -> Union["Message", bool]:
    """\nExample:
            message.edit_text("hello")
    Parameters:
        chat_id (``int`` | ``str``):
            Unique identifier (int) or username (str) of the target chat.
            For your personal cloud (Saved Messages)
            you can simply use "me" or "self".
            For a contact that exists in your Telegram address book
            you can use his phone number (str).
        message_id (``int``):
            Message identifier in the chat specified in chat_id.
        text (``str``):
            New text of the message.
        del_in (``int``):
            Time in Seconds for delete that message.
        parse_mode (:obj:`enums.ParseMode`, *optional*):
            By default, texts are parsed using
            both Markdown and HTML styles.
            You can combine both syntaxes together.
            Pass "markdown" or "md" to enable
            Markdown-style parsing only.
            Pass "html" to enable HTML-style parsing only.
            Pass None to completely disable style parsing.
        entities (List of :obj:`~pyrogram.types.MessageEntity`):
            List of special entities that appear in message text,
            which can be specified instead of *parse_mode*.
        disable_web_page_preview (``bool``, *optional*):
            Disables link previews for links in this message.
        reply_markup (:obj:`InlineKeyboardMarkup`, *optional*):
            An InlineKeyboardMarkup object.
    Returns:
        On success, the edited
        :obj:`Message` or True is returned.
    Raises:
        RPCError: In case of a Telegram RPC error.
    """
    msg = await self.edit_message_text(
        chat_id=chat_id, message_id=message_id, text=text, *args, **kwargs
    )
    if del_in == 0:
        return msg
    await asyncio.sleep(del_in)
    return bool(await msg.delete())


Client.edit_msg_text = edit_message_text



from html import escape
from re import compile as compilere
from re import sub


async def cleanhtml(raw_html: str) -> str:
    """Clean html data."""
    cleanr = compilere("<.*?>")
    return sub(cleanr, "", raw_html)


async def escape_markdown(text: str) -> str:
    """Escape markdown data."""
    escape_chars = r"\*_`\["
    return sub(r"([%s])" % escape_chars, r"\\\1", text)


async def mention_html(name: str, user_id: int) -> str:
    """Mention user in html format."""
    name = escape(name)
    return f'<a href="tg://user?id={user_id}">{name}</a>'


async def mention_markdown(name: str, user_id: int) -> str:
    """Mention user in markdown format."""
    return f"[{(await escape_markdown(name))}](tg://user?id={user_id})"


# Clean File
async def remove_markdown_and_html(text: str) -> str:
    return await clean_markdown(await clean_html(text))


async def clean_html(text: str) -> str:
    return (
        text.replace("<code>", "")
        .replace("</code>", "")
        .replace("<b>", "")
        .replace("</b>", "")
        .replace("<i>", "")
        .replace("</i>", "")
        .replace("<u>", "")
        .replace("</u>", "")
    )


async def clean_markdown(text: str) -> str:
    return text.replace("`", "").replace("**", "").replace("__", "")



from html import escape
from re import compile as compile_re
from time import time
from typing import List

from Abg.helpers.parser import escape_markdown
from pyrogram.enums import ChatType
from pyrogram.types import InlineKeyboardButton, Message

BTN_URL_REGEX = compile_re(r"(\[([^\[]+?)\]\(buttonurl:(?:/{0,2})(.+?)(:same)?\))")

async def parse_button(text: str):
    """Parse button from text."""
    markdown_note = text
    prev = 0
    note_data = ""
    buttons = []
    for match in BTN_URL_REGEX.finditer(markdown_note):
        # Check if btnurl is escaped
        n_escapes = 0
        to_check = match.start(1) - 1
        while to_check > 0 and markdown_note[to_check] == "\\":
            n_escapes += 1
            to_check -= 1

        # if even, not escaped -> create button
        if n_escapes % 2 == 0:
            # create a thruple with button label, url, and newline status
            buttons.append((match.group(2), match.group(3), bool(match.group(4))))
            note_data += markdown_note[prev : match.start(1)]
            prev = match.end(1)
        # if odd, escaped -> move along
        else:
            note_data += markdown_note[prev:to_check]
            prev = match.start(1) - 1
    else:
        note_data += markdown_note[prev:]

    return note_data, buttons


async def build_keyboard(buttons):
    """Build keyboards from provided buttons."""
    keyb = []
    for btn in buttons:
        if btn[-1] and keyb:
            keyb[-1].append(InlineKeyboardButton(btn[0], url=btn[1]))
        else:
            keyb.append([InlineKeyboardButton(btn[0], url=btn[1])])

    return keyb


SMART_OPEN = "“"
SMART_CLOSE = "”"
START_CHAR = ("'", '"', SMART_OPEN)


async def escape_invalid_curly_brackets(text: str, valids: List[str]) -> str:
    new_text = ""
    idx = 0
    while idx < len(text):
        if text[idx] == "{":
            if idx + 1 < len(text) and text[idx + 1] == "{":
                idx += 2
                new_text += "{{{{"
                continue
            success = False
            for v in valids:
                if text[idx:].startswith("{" + v + "}"):
                    success = True
                    break
            if success:
                new_text += text[idx : idx + len(v) + 2]
                idx += len(v) + 2
                continue
            new_text += "{{"

        elif text[idx] == "}":
            if idx + 1 < len(text) and text[idx + 1] == "}":
                idx += 2
                new_text += "}}}}"
                continue
            new_text += "}}"

        else:
            new_text += text[idx]
        idx += 1

    return new_text


async def escape_mentions_using_curly_brackets(
    m: Message,
    text: str,
    parse_words: list,
) -> str:
    teks = await escape_invalid_curly_brackets(text, parse_words)
    if teks:
        teks = teks.format(
            first=escape(m.from_user.first_name),
            last=escape(m.from_user.last_name or m.from_user.first_name),
            mention=m.from_user.mention,
            username=(
                "@" + (await escape_markdown(escape(m.from_user.username)))
                if m.from_user.username
                else m.from_user.mention
            ),
            fullname=" ".join(
                [
                    escape(m.from_user.first_name),
                    escape(m.from_user.last_name),
                ]
                if m.from_user.last_name
                else [escape(m.from_user.first_name)],
            ),
            chatname=escape(m.chat.title)
            if m.chat.type == ChatType.SUPERGROUP
            else escape(m.from_user.first_name),
            id=m.from_user.id,
        )
    else:
        teks = ""

    return teks

def split_limits(text):
    if len(text) < 2048:
        return [text]

    lines = text.splitlines(True)
    small_msg = ""
    result = []
    for line in lines:
        if len(small_msg) + len(line) < 2048:
            small_msg += line
        else:
            result.append(small_msg)
            small_msg = line

    result.append(small_msg)
    return result
    
async def split_quotes(text: str):
    """Split quotes in text."""
    if not any(text.startswith(char) for char in START_CHAR):
        return text.split(None, 1)
    counter = 1  # ignore first char -> is some kind of quote
    while counter < len(text):
        if text[counter] == "\\":
            counter += 1
        elif text[counter] == text[0] or (
            text[0] == SMART_OPEN and text[counter] == SMART_CLOSE
        ):
            break
        counter += 1
    else:
        return text.split(None, 1)

    # 1 to avoid starting quote, and counter is exclusive so avoids ending
    key = await remove_escapes(text[1:counter].strip())
    # index will be in range, or `else` would have been executed and returned
    rest = text[counter + 1 :].strip()
    if not key:
        key = text[0] + text[0]
    return list(filter(None, [key, rest]))


async def remove_escapes(text: str) -> str:
    """Remove the escaped from message."""
    res = ""
    is_escaped = False
    for counter in range(len(text)):
        if is_escaped:
            res += text[counter]
            is_escaped = False
        elif text[counter] == "\\":
            is_escaped = True
        else:
            res += text[counter]
    return res



from enum import IntEnum, unique

from pyrogram import Client
from pyrogram.types import Message


@unique
class Types(IntEnum):
    TEXT = 1
    DOCUMENT = 2
    PHOTO = 3
    VIDEO = 4
    STICKER = 5
    AUDIO = 6
    VOICE = 7
    VIDEO_NOTE = 8
    ANIMATION = 9
    ANIMATED_STICKER = 10
    CONTACT = 11

async def send_cmd(c: Client, msgtype: int):
    GET_FORMAT = {
        Types.TEXT.value: c.send_message,
        Types.DOCUMENT.value: c.send_document,
        Types.PHOTO.value: c.send_photo,
        Types.VIDEO.value: c.send_video,
        Types.STICKER.value: c.send_sticker,
        Types.AUDIO.value: c.send_audio,
        Types.VOICE.value: c.send_voice,
        Types.VIDEO_NOTE.value: c.send_video_note,
        Types.ANIMATION.value: c.send_animation,
        Types.ANIMATED_STICKER.value: c.send_sticker,
        Types.CONTACT: c.send_contact,
    }
    return GET_FORMAT[msgtype]


async def get_note_type(m: Message):
    """Get type of note."""
    if len(m.text.split()) <= 1:
        return None, None, None, None

    data_type = None
    content = None
    raw_text = m.text.markdown if m.text else m.caption.markdown
    args = raw_text.split(None, 2)
    note_name = args[1]

    if len(args) >= 3:
        text = args[2]
        data_type = Types.TEXT

    elif m.reply_to_message:
        if m.reply_to_message.text:
            text = m.reply_to_message.text.markdown
        elif m.reply_to_message.caption:
            text = m.reply_to_message.caption.markdown
        else:
            text = ""

        if len(args) >= 2 and m.reply_to_message.text:  # not caption, text
            data_type = Types.TEXT

        elif m.reply_to_message.sticker:
            content = m.reply_to_message.sticker.file_id
            data_type = Types.STICKER

        elif m.reply_to_message.document:
            if m.reply_to_message.document.mime_type == "application/x-bad-tgsticker":
                data_type = Types.ANIMATED_STICKER
            else:
                data_type = Types.DOCUMENT
            content = m.reply_to_message.document.file_id

        elif m.reply_to_message.photo:
            content = m.reply_to_message.photo.file_id  # last elem = best quality
            data_type = Types.PHOTO

        elif m.reply_to_message.audio:
            content = m.reply_to_message.audio.file_id
            data_type = Types.AUDIO

        elif m.reply_to_message.voice:
            content = m.reply_to_message.voice.file_id
            data_type = Types.VOICE

        elif m.reply_to_message.video:
            content = m.reply_to_message.video.file_id
            data_type = Types.VIDEO

        elif m.reply_to_message.video_note:
            content = m.reply_to_message.video_note.file_id
            data_type = Types.VIDEO_NOTE

        elif m.reply_to_message.animation:
            content = m.reply_to_message.animation.file_id
            data_type = Types.ANIMATION

    else:
        return None, None, None, None

    return note_name, text, data_type, content


async def get_filter_type(m: Message):
    """Get filter type."""
    if len(m.text.split()) <= 1:
        return None, None, None, None

    data_type = None
    content = None
    raw_text = m.text.markdown if m.text else m.caption.markdown
    args = raw_text.split(None, 2)

    if not m.reply_to_message and m.text and len(m.text.split()) >= 3:
        content = None
        text = m.text.split(None, 2)[2]
        data_type = Types.TEXT

    elif m.reply_to_message:
        if m.reply_to_message.text:
            text = m.reply_to_message.text.markdown
        elif m.reply_to_message.caption:
            text = m.reply_to_message.caption.markdown
        else:
            text = ""

        if len(args) >= 2 and m.reply_to_message.text:  # not caption, text
            data_type = Types.TEXT

        elif m.reply_to_message.sticker:
            content = m.reply_to_message.sticker.file_id
            data_type = Types.STICKER

        elif m.reply_to_message.document:
            if m.reply_to_message.document.mime_type == "application/x-bad-tgsticker":
                data_type = Types.ANIMATED_STICKER
            else:
                data_type = Types.DOCUMENT
            content = m.reply_to_message.document.file_id

        elif m.reply_to_message.photo:
            content = m.reply_to_message.photo.file_id  # last elem = best quality
            data_type = Types.PHOTO

        elif m.reply_to_message.audio:
            content = m.reply_to_message.audio.file_id
            data_type = Types.AUDIO

        elif m.reply_to_message.voice:
            content = m.reply_to_message.voice.file_id
            data_type = Types.VOICE

        elif m.reply_to_message.video:
            content = m.reply_to_message.video.file_id
            data_type = Types.VIDEO

        elif m.reply_to_message.video_note:
            content = m.reply_to_message.video_note.file_id
            data_type = Types.VIDEO_NOTE

        elif m.reply_to_message.animation:
            content = m.reply_to_message.animation.file_id
            data_type = Types.ANIMATION

    else:
        text = None
        data_type = None
        content = None

    return text, data_type, content


async def get_wlcm_type(m: Message):
    """Get wlcm type."""
    data_type = None
    content = None
    raw_text = m.text.markdown if m.text else m.caption.markdown
    args = raw_text.split(None, 1)

    if not m.reply_to_message and m.text and len(m.text.split()) >= 2:
        content = None
        text = m.text.split(None, 1)[1]
        data_type = Types.TEXT

    elif m.reply_to_message:
        if m.reply_to_message.text:
            text = m.reply_to_message.text.markdown
        elif m.reply_to_message.caption:
            text = m.reply_to_message.caption.markdown
        else:
            text = ""

        if len(args) >= 1 and m.reply_to_message.text:  # not caption, text
            data_type = Types.TEXT

        elif m.reply_to_message.sticker:
            content = m.reply_to_message.sticker.file_id
            data_type = Types.STICKER

        elif m.reply_to_message.document:
            if m.reply_to_message.document.mime_type == "application/x-bad-tgsticker":
                data_type = Types.ANIMATED_STICKER
            else:
                data_type = Types.DOCUMENT
            content = m.reply_to_message.document.file_id

        elif m.reply_to_message.photo:
            content = m.reply_to_message.photo.file_id  # last elem = best quality
            data_type = Types.PHOTO

        elif m.reply_to_message.audio:
            content = m.reply_to_message.audio.file_id
            data_type = Types.AUDIO

        elif m.reply_to_message.voice:
            content = m.reply_to_message.voice.file_id
            data_type = Types.VOICE

        elif m.reply_to_message.video:
            content = m.reply_to_message.video.file_id
            data_type = Types.VIDEO

        elif m.reply_to_message.video_note:
            content = m.reply_to_message.video_note.file_id
            data_type = Types.VIDEO_NOTE

        elif m.reply_to_message.animation:
            content = m.reply_to_message.animation.file_id
            data_type = Types.ANIMATION

    else:
        text = None
        data_type = None
        content = None

    return text, data_type, content



import html
import io
from asyncio import get_event_loop
from asyncio import sleep as asleep
from logging import getLogger
from typing import Union

from pyrogram.errors import (
    ChatAdminRequired,
    ChatWriteForbidden,
    FloodWait,
    MessageAuthorRequired,
    MessageDeleteForbidden,
    MessageIdInvalid,
    MessageNotModified,
    MessageTooLong,
)
from pyrogram.types import Message

LOGGER = getLogger(__name__)

Message.input = property(
    lambda m: m.text[m.text.find(m.command[0]) + len(m.command[0]) + 1 :]
    if len(m.command) > 1
    else None
)


async def reply_text(
    self: Message, text: str, as_raw: bool = False, del_in: int = 0, *args, **kwargs
) -> Union["Message", bool]:
    """\nExample:
            message.reply_msg("hello")
    Parameters:
        text (``str``):
            Text of the message to be sent.
        del_in (``int``):
            Time in Seconds for delete that message.
        quote (``bool``, *optional*):
            If ``True``, the message will be sent as
            a reply to this message.
            If *reply_to_message_id* is passed,
            this parameter will be ignored.
            Defaults to ``True`` in group chats
            and ``False`` in private chats.
        parse_mode (:obj:`enums.ParseMode`, *optional*):
            By default, texts are parsed using both
            Markdown and HTML styles.
            You can combine both syntaxes together.
            Pass "markdown" or "md" to enable
            Markdown-style parsing only.
            Pass "html" to enable HTML-style parsing only.
            Pass None to completely disable style parsing.
        disable_web_page_preview (``bool``, *optional*):
            Disables link previews for links in this message.
        disable_notification (``bool``, *optional*):
            Sends the message silently.
            Users will receive a notification with no sound.
        reply_to_message_id (``int``, *optional*):
            If the message is a reply, ID of the original message.
        schedule_date (:py:obj:`~datetime.datetime`, *optional*):
            Date when the message will be automatically sent. Unix time.
        protect_content (``bool``, *optional*):
            Protects the contents of the sent message from forwarding and saving.
        reply_markup (:obj:`InlineKeyboardMarkup`
        | :obj:`ReplyKeyboardMarkup` | :obj:`ReplyKeyboardRemove`
        | :obj:`ForceReply`, *optional*):
            Additional interface options. An object for an inline keyboard,
            custom reply keyboard,
            instructions to remove reply keyboard or to
            force a reply from the user.
    Returns:
        On success, the sent Message or True is returned.
    Raises:
        RPCError: In case of a Telegram RPC error.
    """
    try:
        if as_raw:
            msg = await self.reply_text(
                text=f"<code>{html.escape(text.html)}</code>", *args, **kwargs
            )
        else:
            msg = await self.reply_text(text=text, *args, **kwargs)
        if del_in == 0:
            return msg
        await asleep(del_in)
        return bool(await msg.delete_msg())
    except FloodWait as e:
        await asleep(e.value)
        return await reply_text(self, text, *args, **kwargs)
    except (ChatWriteForbidden, ChatAdminRequired):
        LOGGER.info(
            f"Leaving from {self.chat.title} [{self.chat.id}] because doesn't have admin permission."
        )
        return await self.chat.leave()


async def edit_text(
    self, text: str, del_in: int = 0, *args, **kwargs
) -> Union["Message", bool]:
    """\nExample:
            message.edit_msg("hello")
    Parameters:
        text (``str``):
            New text of the message.
        del_in (``int``):
            Time in Seconds for delete that message.
        parse_mode (:obj:`enums.ParseMode`, *optional*):
            By default, texts are parsed using
            both Markdown and HTML styles.
            You can combine both syntaxes together.
            Pass "markdown" or "md" to enable
            Markdown-style parsing only.
            Pass "html" to enable HTML-style parsing only.
            Pass None to completely disable style parsing.
        disable_web_page_preview (``bool``, *optional*):
            Disables link previews for links in this message.
        reply_markup (:obj:`InlineKeyboardMarkup`, *optional*):
            An InlineKeyboardMarkup object.
    Returns:
        On success, the edited
        :obj:`Message` or True is returned.
    Raises:
        RPCError: In case of a Telegram RPC error.
    """
    try:
        msg = await self.edit_text(text, *args, **kwargs)
        if del_in == 0:
            return msg
        await asleep(del_in)
        return bool(await msg.delete_msg())
    except FloodWait as e:
        LOGGER.warning(str(e))
        await asleep(e.value)
        return await edit_text(self, text, *args, **kwargs)
    except MessageNotModified:
        return False
    except (ChatWriteForbidden, ChatAdminRequired):
        LOGGER.info(
            f"Leaving from {self.chat.title} [{self.chat.id}] because doesn't have admin permission."
        )
        return await self.chat.leave()
    except (MessageAuthorRequired, MessageIdInvalid):
        return await reply_text(self, text=text, *args, **kwargs)


async def edit_or_send_as_file(
    self, text: str, del_in: int = 0, as_raw: bool = False, *args, **kwargs
) -> Union["Message", bool]:
    """\nThis will first try to message.edit.
    If it raises MessageTooLong error,
    run message.send_as_file.
    Example:
            message.edit_or_send_as_file("some huge text")
    Parameters:
        text (``str``):
            New text of the message.
        del_in (``int``):
            Time in Seconds for delete that message.
        log (``bool`` | ``str``, *optional*):
            If ``True``, the message will be forwarded
            to the log channel.
            If ``str``, the logger name will be updated.
        sudo (``bool``, *optional*):
            If ``True``, sudo users supported.
        as_raw (``bool``, *optional*):
            If ``False``, the message will be escaped with current parse mode.
            default to ``False``.
        parse_mode (:obj:`enums.ParseMode`, *optional*):
            By default, texts are parsed using both
            Markdown and HTML styles.
            You can combine both syntaxes together.
            Pass "markdown" or "md" to enable
            Markdown-style parsing only.
            Pass "html" to enable HTML-style parsing only.
            Pass None to completely disable style parsing.
        disable_web_page_preview (``bool``, *optional*):
            Disables link previews for links in this message.
        reply_markup (:obj:`InlineKeyboardMarkup`, *optional*):
            An InlineKeyboardMarkup object.
        **kwargs (for message.send_as_file)
    Returns:
        On success, the edited
        :obj:`Message` or True is returned.
    Raises:
        RPCError: In case of a Telegram RPC error.
    """
    text = html.escape(text.html) if as_raw else text
    try:
        msg = await edit_text(self, text=text, *args, **kwargs)
        if del_in == 0:
            return msg
        await asleep(del_in)
        return bool(await msg.delete_msg())
    except (MessageTooLong, OSError):
        return await reply_as_file(self, text=text, *args, **kwargs)


async def reply_or_send_as_file(
    self, text: str, as_raw: bool = False, del_in: int = 0, *args, **kwargs
) -> Union["Message", bool]:
    """\nThis will first try to message.reply.
    If it raise MessageTooLong error,
    run message.send_as_file.
    Example:
            message.reply_or_send_as_file("some huge text")
    Parameters:
        text (``str``):
            Text of the message to be sent.
        del_in (``int``):
            Time in Seconds for delete that message.
        quote (``bool``, *optional*):
            If ``True``, the message will be sent
            as a reply to this message.
            If *reply_to_message_id* is passed,
            this parameter will be ignored.
            Defaults to ``True`` in group chats
            and ``False`` in private chats.
        parse_mode (:obj:`enums.ParseMode`, *optional*):
            By default, texts are parsed using
            both Markdown and HTML styles.
            You can combine both syntaxes together.
            Pass "markdown" or "md" to enable
            Markdown-style parsing only.
            Pass "html" to enable HTML-style parsing only.
            Pass None to completely disable style parsing.
        disable_web_page_preview (``bool``, *optional*):
            Disables link previews for links in this message.
        disable_notification (``bool``, *optional*):
            Sends the message silently.
            Users will receive a notification with no sound.
        reply_to_message_id (``int``, *optional*):
            If the message is a reply, ID of the
            original message.
        reply_markup (:obj:`InlineKeyboardMarkup`
        | :obj:`ReplyKeyboardMarkup` | :obj:`ReplyKeyboardRemove`
        | :obj:`ForceReply`, *optional*):
            Additional interface options. An object for an
            inline keyboard, custom reply keyboard,
            instructions to remove reply keyboard
            or to force a reply from the user.
        **kwargs (for message.send_as_file)
    Returns:
        On success, the sent Message or True is returned.
    Raises:
        RPCError: In case of a Telegram RPC error.
    """
    text = html.escape(text.html) if as_raw else text
    try:
        return await reply_text(self, text=text, del_in=del_in, *args, **kwargs)
    except MessageTooLong:
        return await reply_as_file(self, text=text, **kwargs)


async def reply_as_file(
    self,
    text: str,
    filename: str = "output.txt",
    caption: str = "",
    delete_message: bool = True,
):
    """\nYou can send large outputs as file
    Example:
        message.reply_as_file(text="hello")
    Parameters:
        text (``str``):
            Text of the message to be sent.
        filename (``str``, *optional*):
            file_name for output file.
        caption (``str``, *optional*):
            caption for output file.
        delete_message (``bool``, *optional*):
            If ``True``, the message will be deleted
            after sending the file.
    Returns:
        On success, the sent Message is returned.
    """
    reply_to_id = self.reply_to_message.id if self.reply_to_message else self.id
    if delete_message:
        get_event_loop().create_task(self.delete())
    doc = io.BytesIO(text.encode())
    doc.name = filename
    return await self.reply_document(
        document=doc,
        caption=caption[:1024],
        disable_notification=True,
        reply_to_message_id=reply_to_id,
    )


async def delete(self, revoke: bool = True) -> bool:
    """\nThis will first try to delete and ignore
    it if it raises MessageDeleteForbidden
    Parameters:
        revoke (``bool``, *optional*):
            Deletes messages on both parts.
            This is only for private cloud chats and normal groups, messages on
            channels and supergroups are always revoked (i.e.: deleted for everyone).
            Defaults to True.
    Returns:
        True on success, False otherwise.
    """
    try:
        return bool(await self.delete(revoke=revoke))
    except FloodWait as e:
        LOGGER.warning(str(e))
        await asleep(e.value)
        return await delete(self, revoke)
    except MessageDeleteForbidden:
        return False
    except Exception as e:
        LOGGER.warning(str(e))


Message.reply_msg = reply_text
Message.edit_msg = edit_text
Message.edit_or_send_as_file = edit_or_send_as_file
Message.reply_or_send_as_file = reply_or_send_as_file
Message.reply_as_file = reply_as_file
Message.delete_msg = delete



from functools import wraps
from typing import Callable, Union

from cachetools import TTLCache
from pyrate_limiter import (
    BucketFullException,
    Duration,
    Limiter,
    MemoryListBucket,
    RequestRate,
)
from pyrogram import Client
from pyrogram.errors import QueryIdInvalid
from pyrogram.types import CallbackQuery, Message

# you need to install this using : pip3 install pyrate_limiter


class RateLimiter:
    """
    Implement rate limit logic using leaky bucket
    algorithm, via pyrate_limiter.
    (https://pypi.org/project/pyrate-limiter/)
    """

    def __init__(self) -> None:
        # 2 requests per seconds
        self.second_rate = RequestRate(2, Duration.SECOND)

        # 15 requests per minute.
        self.minute_rate = RequestRate(15, Duration.MINUTE)

        # 500 requests per hour
        self.hourly_rate = RequestRate(500, Duration.HOUR)

        # 1500 requests per day
        self.daily_rate = RequestRate(1500, Duration.DAY)

        self.limiter = Limiter(
            self.minute_rate,
            self.hourly_rate,
            self.daily_rate,
            bucket_class=MemoryListBucket,
        )

    async def acquire(self, userid: Union[int, str]) -> bool:
        """
        Acquire rate limit per userid and return True / False
        based on userid ratelimit status.
        """

        try:
            self.limiter.try_acquire(userid)
            return False
        except BucketFullException:
            return True


ratelimit = RateLimiter()

warned_users = TTLCache(maxsize=128, ttl=60)

warning_message = "sᴘᴀᴍ ᴅᴇᴛᴇᴄᴛᴇᴅ! ɪɢɴᴏʀɪɴɢ ʏᴏᴜʀ ᴀʟʟ ʀᴇǫᴜᴇsᴛs ғᴏʀ ғᴇᴡ ᴍɪɴᴜᴛᴇs."


def ratelimiter(func: Callable) -> Callable:
    """
    Restricts user's from spamming commands or pressing buttons multiple times
    using leaky bucket algorithm and pyrate_limiter.
    """

    @wraps(func)
    async def decorator(client: Client, update: Union[Message, CallbackQuery]):
        userid = update.from_user.id if update.from_user else update.sender_chat.id
        is_limited = await ratelimit.acquire(userid)

        if is_limited and userid not in warned_users:
            if isinstance(update, Message):
                await update.reply_text(warning_message)
                warned_users[userid] = 1
                return

            elif isinstance(update, CallbackQuery):
                try:
                    await update.answer(warning_message, show_alert=True)
                except QueryIdInvalid:
                    warned_users[userid] = 1
                    return
                warned_users[userid] = 1
                return

        elif is_limited and userid in warned_users:
            pass
        else:
            return await func(client, update)

    return decorator



import asyncio
import math
import time

from pyrogram.errors import FloodWait, MessageIdInvalid, MessageNotModified


async def progress_for_pyrogram(current, total, ud_type, message, start, dc_id):
    """generic progress display for Telegram Upload / Download status"""
    now = time.time()
    diff = now - start
    if round(diff % 10.00) == 0 or current == total:
        # if round(current / total * 100, 0) % 5 == 0:
        percentage = current * 100 / total
        elapsed_time = round(diff)
        if elapsed_time == 0:
            return
        speed = current / diff
        time_to_completion = round((total - current) / speed)
        estimated_total_time = elapsed_time + time_to_completion

        elapsed_time = time_formatter(elapsed_time)
        estimated_total_time = time_formatter(estimated_total_time)

        progress = "[{0}{1}] \nP: {2}%\n".format(
            "".join(["█" for _ in range(math.floor(percentage / 5))]),
            "".join(["░" for _ in range(20 - math.floor(percentage / 5))]),
            round(percentage, 2),
        )

        tmp = (
            progress
            + "{0} <b>of</b> {1}\n<b>sᴘᴇᴇᴅ:</b> {2}/s\n<b>ᴅᴄ ɪᴅ=:</b> {3}\n<b>ᴇᴛᴀ:</b> {4}</b>\n".format(
                humanbytes(current),
                humanbytes(total),
                humanbytes(speed),
                dc_id,
                estimated_total_time if estimated_total_time != "" else "0 s",
            )
        )
        try:
            await message.edit(f"{ud_type}\n {tmp}")
        except FloodWait as e:
            await asyncio.sleep(e.value)
            await message.edit(f"{ud_type}\n {tmp}")
        except (MessageNotModified, MessageIdInvalid):
            pass


def humanbytes(size: int) -> str:
    """converts bytes into human readable format"""
    # https://stackoverflow.com/a/49361727/4723940
    # 2**10 = 1024
    if not size:
        return ""
    power = 2**10
    number = 0
    dict_power_n = {0: " ", 1: "Ki", 2: "Mi", 3: "Gi", 4: "Ti"}
    while size > power:
        size /= power
        number += 1
    return f"{str(round(size, 2))} {dict_power_n[number]}B"


def time_formatter(seconds: int) -> str:
    result = ""
    v_m = 0
    remainder = seconds
    r_ange_s = {"days": (24 * 60 * 60), "hours": (60 * 60), "minutes": 60, "seconds": 1}
    for age, divisor in r_ange_s.items():
        v_m, remainder = divmod(remainder, divisor)
        v_m = int(v_m)
        if v_m != 0:
            result += f" {v_m} {age} "
    return result



class PyromodConfig:
    timeout_handler = None
    stopped_handler = None
    throw_exceptions = True
    unallowed_click_alert = True
    unallowed_click_alert_text = "ʏᴏᴜ'ʀᴇ ɴᴏᴛ ᴀᴜᴛʜᴏʀɪᴢᴇᴅ ᴛᴏ ᴄʟɪᴄᴋ ᴛʜɪs ʙᴜᴛᴛᴏɴ."


def patch(obj):
    def is_patchable(item):
        return getattr(item[1], "patchable", False)

    def wrapper(container):
        for name, func in filter(is_patchable, container.__dict__.items()):
            old = getattr(obj, name, None)
            if old is not None:  # Not adding 'old' to new func
                setattr(obj, f"old{name}", old)
            setattr(obj, name, func)
        return container

    return wrapper


def patchable(func):
    func.patchable = True
    return func



from re import compile as compile_re
from re import escape
from shlex import split
from typing import List, Union

from pyrogram import Client
from pyrogram.enums import ChatMemberStatus as CMS
from pyrogram.enums import ChatType
from pyrogram.errors import UserNotParticipant
from pyrogram.filters import create
from pyrogram.types import Message

PREFIX_HANDLER = ["/", "!", "~", ".", "+", "*"]


def command(
    commands: Union[str, List[str]],
    case_sensitive: bool = False,
):
    async def func(flt, c: Client, m: Message):
        if not m:
            return

        date = m.edit_date
        if date:
            return  # reaction

        if m.chat and m.chat.type == ChatType.CHANNEL:
            return

        if m and not m.from_user:
            return False

        if m.from_user.is_bot:
            return False

        if any([m.forward_from_chat, m.forward_from]):
            return False

        text: str = m.text or m.caption
        if not text:
            return False
        regex = r"^[{prefix}](\w+)(@{bot_name})?(?: |$)(.*)".format(
            prefix="|".join(escape(x) for x in PREFIX_HANDLER),
            bot_name=c.me.username,
        )
        matches = compile_re(regex).search(text)
        if matches:
            m.command = [matches.group(1)]
            if matches.group(1) not in flt.commands:
                return False
            if bool(m.chat and m.chat.type in {ChatType.SUPERGROUP}):
                try:
                    user_status = (await m.chat.get_member(m.from_user.id)).status
                except UserNotParticipant:
                    # i.e anon admin
                    user_status = CMS.ADMINISTRATOR
                except ValueError:
                    # i.e. PM
                    user_status = CMS.OWNER
            if matches.group(3) == "":
                return True
            try:
                for arg in split(matches.group(3)):
                    m.command.append(arg)
            except ValueError:
                pass
            return True
        return False

    commands = commands if isinstance(commands, list) else [commands]
    commands = {c if case_sensitive else c.lower() for c in commands}

    return create(
        func,
        "NormalCommandFilter",
        commands=commands,
        case_sensitive=case_sensitive,
    )



import asyncio
from typing import Union

from pyrogram import Client
from pyrogram.types import Message


async def send_message(
    self, chat_id: Union[int, str], text: str, del_in: int = 0, *args, **kwargs
) -> Union["Message", bool]:
    """\nSend text messages.
    Example:
            @userge.send_message(chat_id=12345, text='test')
    Parameters:
        chat_id (``int`` | ``str``):
            Unique identifier (int) or username (str) of the target chat.
            For your personal cloud (Saved Messages)
            you can simply use "me" or "self".
            For a contact that exists in your Telegram address book
            you can use his phone number (str).
        text (``str``):
            Text of the message to be sent.
        del_in (``int``):
            Time in Seconds for delete that message.
        log (``bool`` | ``str``, *optional*):
            If ``True``, the message will be forwarded to the log channel.
            If ``str``, the logger name will be updated.
        parse_mode (:obj:`enums.ParseMode`, *optional*):
            By default, texts are parsed using both Markdown and HTML styles.
            You can combine both syntaxes together.
            Pass "markdown" or "md" to enable Markdown-style parsing only.
            Pass "html" to enable HTML-style parsing only.
            Pass None to completely disable style parsing.
        entities (List of :obj:`~pyrogram.types.MessageEntity`):
            List of special entities that appear in message text,
            which can be specified instead of *parse_mode*.
        disable_web_page_preview (``bool``, *optional*):
            Disables link previews for links in this message.
        disable_notification (``bool``, *optional*):
            Sends the message silently.
            Users will receive a notification with no sound.
        reply_to_message_id (``int``, *optional*):
            If the message is a reply, ID of the original message.
        schedule_date (:py:obj:`~datetime.datetime`, *optional*):
            Date when the message will be automatically sent. Unix time.
        protect_content (``bool``, *optional*):
            Protects the contents of the sent message from forwarding and saving.
        reply_markup (:obj:`InlineKeyboardMarkup` | :obj:`ReplyKeyboardMarkup`
        | :obj:`ReplyKeyboardRemove` | :obj:`ForceReply`, *optional*):
            Additional interface options. An object for an inline keyboard,
            custom reply keyboard, instructions to remove
            reply keyboard or to force a reply from the user.
    Returns:
        :obj:`Message`: On success, the sent text message or True is returned.
    """
    msg = await self.send_message(chat_id=chat_id, text=text, *args, **kwargs)
    if del_in == 0:
        return msg
    if del_in > 0:
        await asyncio.sleep(del_in)
        return bool(await msg.delete())


Client.send_msg = send_message



"""
Copyright (C) 2023 Abishnoi69. <https://github.com/Abishnoi69>
This file is part of Abg.
Abg is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Abg is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.
You should have received a copy of the GNU General Public License.

along with Abg.  If not, see <https://www.gnu.org/licenses/>.
"""
from .helpers import *
from .human_read import *
from .parser import *
from .pyro_progress import *
from .msg_types import *
from .string import *



import time

GAP = {}

SIZE_UNITS = ["B", "KB", "MB", "GB", "TB", "PB"]


def get_readable_time(seconds: int) -> str:
    count = 0
    ping_time = ""
    time_list = []
    time_suffix_list = ["sᴇᴄ", "ᴍɪɴ", "ʜᴏᴜʀ", "ᴅᴀʏ", "ᴡᴇᴇᴋ", "ᴍᴏɴᴛʜ", "ʏᴇᴀʀ"]
    while count < 4:
        count += 1
        remainder, result = divmod(seconds, 60) if count < 3 else divmod(seconds, 24)
        if seconds == 0 and remainder == 0:
            break
        time_list.append(int(result))
        seconds = int(remainder)
    for i in range(len(time_list)):
        time_list[i] = str(time_list[i]) + time_suffix_list[i]
    if len(time_list) == 4:
        ping_time += f"{time_list.pop()}, "
    time_list.reverse()
    ping_time += ":".join(time_list)
    return ping_time


async def convert_seconds_to_minutes(seconds: int):
    seconds = int(seconds)
    seconds %= 24 * 3600
    seconds %= 3600
    minutes = seconds // 60
    seconds %= 60
    return "%02d:%02d" % (minutes, seconds)


def get_readable_time2(seconds: int) -> str:
    result = ""
    (days, remainder) = divmod(seconds, 86400)
    days = int(days)
    if days != 0:
        result += f"{days} ᴅᴀʏ "
    (hours, remainder) = divmod(remainder, 3600)
    hours = int(hours)
    if hours != 0:
        result += f"{hours} ʜᴏᴜʀ "
    (minutes, seconds) = divmod(remainder, 60)
    minutes = int(minutes)
    if minutes != 0:
        result += f"{minutes} ᴍɪɴᴜᴛᴇ "
    seconds = int(seconds)
    result += f"{seconds} sᴇᴄᴏɴᴅ "
    return result


def get_readable_file_size(size_in_bytes) -> str:
    if size_in_bytes is None:
        return "0B"
    index = 0
    while size_in_bytes >= 1024:
        size_in_bytes /= 1024
        index += 1
    try:
        return f"{round(size_in_bytes, 2)}{SIZE_UNITS[index]}"
    except IndexError:
        return "File too large"


def get_readable_bitrate(bitrate_kbps):
    if bitrate_kbps > 10000:
        bitrate = str(round(bitrate_kbps / 1000, 2)) + " " + "ᴍʙ/s"
    else:
        bitrate = str(round(bitrate_kbps, 2)) + " " + "ᴋʙ/s"

    return bitrate


async def check_time_gap(user_id: int):
    """A Function for checking user time gap!
    :parameter user_id Telegram User ID"""

    if str(user_id) in GAP:
        current_time = time.time()
        previous_time = GAP[str(user_id)]
        if round(current_time - previous_time) < 10:
            return True, round(previous_time - current_time + 10)
        del GAP[str(user_id)]
    else:
        GAP[str(user_id)] = time.time()
    return False, None



"""
thanks 
__author__ = 'PyMaster'
"""

from .inline_keyboard import InlineButton, InlineKeyboard
from .inline_pagination_keyboard import InlineButton, InlinePaginationKeyboard
from .reply_keyboard import ForceReply, ReplyButton, ReplyKeyboard, ReplyKeyboardRemove



from asyncio import gather

import httpx
from aiohttp import ClientSession

# Aiohttp Async Client
aiohttpsession = ClientSession()

# HTTPx Async Client
http = httpx.AsyncClient(
    http2=True,
    verify=False,
    timeout=httpx.Timeout(40),
)


async def get(url: str, *args, **kwargs):
    async with aiohttpsession.get(url, *args, **kwargs) as resp:
        try:
            data = await resp.json()
        except Exception:
            data = await resp.text()
    return data


async def head(url: str, *args, **kwargs):
    async with aiohttpsession.head(url, *args, **kwargs) as resp:
        try:
            data = await resp.json()
        except Exception:
            data = await resp.text()
    return data


async def post(url: str, *args, **kwargs):
    async with aiohttpsession.post(url, *args, **kwargs) as resp:
        try:
            data = await resp.json()
        except Exception:
            data = await resp.text()
    return data


async def multiget(url: str, times: int, *args, **kwargs):
    return await gather(*[get(url, *args, **kwargs) for _ in range(times)])


async def multihead(url: str, times: int, *args, **kwargs):
    return await gather(*[head(url, *args, **kwargs) for _ in range(times)])


async def multipost(url: str, times: int, *args, **kwargs):
    return await gather(*[post(url, *args, **kwargs) for _ in range(times)])


async def resp_get(url: str, *args, **kwargs):
    return await aiohttpsession.get(url, *args, **kwargs)


async def resp_post(url: str, *args, **kwargs):
    return await aiohttpsession.post(url, *args, **kwargs)



from typing import List, Union

from pyrogram.emoji import *
from pyrogram.types import InlineKeyboardButton, InlineKeyboardMarkup


class InlineKeyboard(InlineKeyboardMarkup):
    _SYMBOL_FIRST_PAGE = "« {}"
    _SYMBOL_PREVIOUS_PAGE = "‹ {}"
    _SYMBOL_CURRENT_PAGE = "· {} ·"
    _SYMBOL_NEXT_PAGE = "{} ›"
    _SYMBOL_LAST_PAGE = "{} »"
    _LOCALES = {
        "be_BY": f"{FLAG_BELARUS} Беларуская",  # Belarusian - Belarus
        "de_DE": f"{FLAG_GERMANY} Deutsch",  # German - Germany
        "zh_CN": f"{FLAG_CHINA} 中文",  # Chinese - China
        # English - United States
        "en_US": f"{FLAG_UNITED_KINGDOM}  English",
        "fr_FR": f"{FLAG_FRANCE} Français",  # French - France
        # Indonesian - Indonesia
        "id_ID": f"{FLAG_INDONESIA} Bahasa Indonesia",
        "it_IT": f"{FLAG_ITALY} Italiano",  # Italian - Italy
        "ko_KR": f"{FLAG_SOUTH_KOREA} 한국어",  # Korean - Korea
        "tr_TR": f"{FLAG_TURKEY} Türkçe",  # Turkish - Turkey
        "ru_RU": f"{FLAG_RUSSIA} Русский",  # Russian - Russia
        "es_ES": f"{FLAG_SPAIN} Español",  # Spanish - Spain
        "uk_UA": f"{FLAG_UKRAINE} Українська",  # Ukrainian - Ukraine
        "uz_UZ": f"{FLAG_UZBEKISTAN} Oʻzbekcha",  # Uzbek - Uzbekistan
    }

    def __init__(self, row_width=3):
        self.inline_keyboard = list()
        super().__init__(inline_keyboard=self.inline_keyboard)
        self.row_width = row_width

    def add(self, *args):
        self.inline_keyboard = [
            args[i : i + self.row_width] for i in range(0, len(args), self.row_width)
        ]

    def row(self, *args):
        self.inline_keyboard.append([button for button in args])

    def _add_button(self, text, callback_data):
        return InlineKeyboardButton(
            text=text, callback_data=self.callback_pattern.format(number=callback_data)
        )

    @property
    def _left_pagination(self):
        return [
            self._add_button(self._SYMBOL_CURRENT_PAGE.format(number), number)
            if number == self.current_page
            else self._add_button(self._SYMBOL_NEXT_PAGE.format(number), number)
            if number == 4
            else self._add_button(
                self._SYMBOL_LAST_PAGE.format(self.count_pages), self.count_pages
            )
            if number == 5
            else self._add_button(number, number)
            for number in range(1, 6)
        ]

    @property
    def _middle_pagination(self):
        return [
            self._add_button(self._SYMBOL_FIRST_PAGE.format(1), 1),
            self._add_button(
                self._SYMBOL_PREVIOUS_PAGE.format(self.current_page - 1),
                self.current_page - 1,
            ),
            self._add_button(
                self._SYMBOL_CURRENT_PAGE.format(self.current_page), self.current_page
            ),
            self._add_button(
                self._SYMBOL_NEXT_PAGE.format(self.current_page + 1),
                self.current_page + 1,
            ),
            self._add_button(
                self._SYMBOL_LAST_PAGE.format(self.count_pages), self.count_pages
            ),
        ]

    @property
    def _right_pagination(self):
        return [
            self._add_button(self._SYMBOL_FIRST_PAGE.format(1), 1),
            self._add_button(
                self._SYMBOL_PREVIOUS_PAGE.format(self.count_pages - 3),
                self.count_pages - 3,
            ),
        ] + [
            self._add_button(self._SYMBOL_CURRENT_PAGE.format(number), number)
            if number == self.current_page
            else self._add_button(number, number)
            for number in range(self.count_pages - 2, self.count_pages + 1)
        ]

    @property
    def _full_pagination(self):
        return [
            self._add_button(number, number)
            if number != self.current_page
            else self._add_button(self._SYMBOL_CURRENT_PAGE.format(number), number)
            for number in range(1, self.count_pages + 1)
        ]

    @property
    def _build_pagination(self):
        if self.count_pages <= 5:
            return self._full_pagination
        else:
            if self.current_page <= 3:
                return self._left_pagination
            elif self.current_page > self.count_pages - 3:
                return self._right_pagination
            else:
                return self._middle_pagination

    def paginate(self, count_pages: int, current_page: int, callback_pattern: str):
        self.count_pages = count_pages
        self.current_page = current_page
        self.callback_pattern = callback_pattern

        return self.inline_keyboard.append(self._build_pagination)

    def languages(
        self, callback_pattern: str, locales: Union[str, List[str]], row_width: int = 2
    ):
        locales = locales if isinstance(locales, list) else [locales]

        buttons = [
            InlineKeyboardButton(
                text=self._LOCALES.get(locales[i], "Invalid locale"),
                callback_data=callback_pattern.format(locale=locales[i]),
            )
            for i in range(0, len(locales))
        ]

        self.inline_keyboard = [
            buttons[i : i + row_width] for i in range(0, len(buttons), row_width)
        ]


class InlineButton(InlineKeyboardButton):
    def __init__(
        self,
        text=None,
        callback_data=None,
        url=None,
        login_url=None,
        user_id=None,
        switch_inline_query=None,
        switch_inline_query_current_chat=None,
        callback_game=None,
    ):
        super().__init__(
            text=text,
            callback_data=callback_data,
            url=url,
            login_url=login_url,
            user_id=user_id,
            switch_inline_query=switch_inline_query,
            switch_inline_query_current_chat=switch_inline_query_current_chat,
            callback_game=callback_game,
        )




#  Copyright (C) 2023-present Abishnoi6 <https://github.com/Abishnoi69>
#
#  This file is part of Abg.
#
#  Abg is free software: you can redistribute it and/or modify
#  it under the terms of the GNU Lesser General Public License as published
#  by the Free Software Foundation, either version 3 of the License, or
#  (at your option) any later version.
#
#  Abg is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU Lesser General Public License for more details.
#
#  You should have received a copy of the GNU Lesser General Public License
#  along with Abg.  If not, see <http://www.gnu.org/licenses/>.

# ===================================================================

import setuptools

with open("README.md", encoding="utf8") as readme:
    long_description = readme.read()

"""    
def read(file: str) -> list:
    with open(file, encoding="utf-8") as r:
        return [i.strip() for i in r]
"""

setuptools.setup(
    name="Abg",
    packages=setuptools.find_packages(),
    version="2.1",
    description="Telegram bot helpers",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/Abishnoi69",
    download_url="https://github.com/Abishnoi69/Abg/releases/latest",
    author="Abishnoi1M",
    author_email="Abishnoi69@Abg.org",
    license="MIT",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: Implementation",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Internet",
        "Topic :: Communications",
        "Topic :: Communications :: Chat",
        "Topic :: Software Development :: Libraries",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "Topic :: Software Development :: Libraries :: Application Frameworks",
    ],
    keywords="telegram bot chat messenger mtproto api client library python conversation keyboard userbot patch botapi https",
    project_urls={
        "Tracker": "https://github.com/Abishnoi69/Abg/issues",
        "Community": "https://t.me/Abgpy",
        "Source": "https://github.com/Abishnoi69/Abg",
        "Documentation": "https://github.com/Abishnoi69/Abg/tree/master/doce",
    },
    python_requires="~=3.7",
    # install_requires=read("requirements.txt")
)



import asyncio
from enum import Enum
from inspect import iscoroutinefunction
from typing import Callable, Optional, Union

import pyrogram
from pyrogram.errors import QueryIdInvalid

from .utils import PyromodConfig, patch, patchable


class ListenerStopped(Exception):
    pass


class ListenerTimeout(Exception):
    pass


class ListenerTypes(Enum):
    MESSAGE = "message"
    CALLBACK_QUERY = "callback_query"


@patch(pyrogram.client.Client)
class Client:
    @patchable
    def __init__(self, *args, **kwargs):
        self.listeners = {listener_type: {} for listener_type in ListenerTypes}
        self.old__init__(*args, **kwargs)

    @patchable
    async def listen(
        self,
        identifier: tuple,
        filters=None,
        listener_type=ListenerTypes.MESSAGE,
        timeout=None,
        unallowed_click_alert=True,
    ):
        if type(listener_type) != ListenerTypes:
            raise TypeError(
                "Parameter listener_type should be a"
                " value from pyromod.listen.ListenerTypes"
            )

        future = self.loop.create_future()
        future.add_done_callback(
            lambda f: self.stop_listening(identifier, listener_type)
        )

        listener_data = {
            "future": future,
            "filters": filters,
            "unallowed_click_alert": unallowed_click_alert,
        }

        self.listeners[listener_type].update({identifier: listener_data})

        try:
            return await asyncio.wait_for(future, timeout)
        except asyncio.exceptions.TimeoutError:
            if callable(PyromodConfig.timeout_handler):
                PyromodConfig.timeout_handler(identifier, listener_data, timeout)
            elif PyromodConfig.throw_exceptions:
                raise ListenerTimeout(timeout)

    @patchable
    async def ask(
        self,
        text,
        identifier: tuple,
        filters=None,
        listener_type=ListenerTypes.MESSAGE,
        timeout=None,
        *args,
        **kwargs,
    ):
        request = await self.send_message(identifier[0], text, *args, **kwargs)
        response = await self.listen(identifier, filters, listener_type, timeout)
        if response:
            response.request = request

        return response

    """
    needed for matching when message_id or
    user_id is null, and to take precedence
    """

    @patchable
    def match_listener(
        self,
        data: Optional[tuple] = None,
        listener_type: ListenerTypes = ListenerTypes.MESSAGE,
        identifier_pattern: Optional[tuple] = None,
    ) -> tuple:
        if data:
            listeners = self.listeners[listener_type]
            # case with 3 args on identifier
            # most probably waiting for a specific user
            # to click a button in a specific message
            if data in listeners:
                return listeners[data], data

            # cases with 2 args on identifier
            # (None, user, message) does not make
            # sense since the message_id is not unique
            elif (data[0], data[1], None) in listeners:
                matched = (data[0], data[1], None)
            elif (data[0], None, data[2]) in listeners:
                matched = (data[0], None, data[2])

            # cases with 1 arg on identifier
            # (None, None, message) does not make sense as well
            elif (data[0], None, None) in listeners:
                matched = (data[0], None, None)
            elif (None, data[1], None) in listeners:
                matched = (None, data[1], None)
            else:
                return None, None

            return listeners[matched], matched
        elif identifier_pattern:

            def match_identifier(pattern, identifier):
                comparison = (
                    pattern[0] in (identifier[0], None),
                    pattern[1] in (identifier[1], None),
                    pattern[2] in (identifier[2], None),
                )
                return comparison == (True, True, True)

            for identifier, listener in self.listeners[listener_type].items():
                if match_identifier(identifier_pattern, identifier):
                    return listener, identifier
            return None, None

    @patchable
    def stop_listening(
        self,
        data: Optional[tuple] = None,
        listener_type: ListenerTypes = ListenerTypes.MESSAGE,
        identifier_pattern: Optional[tuple] = None,
    ):
        listener, identifier = self.match_listener(
            data, listener_type, identifier_pattern
        )

        if not listener:
            return
        elif listener["future"].done():
            del self.listeners[listener_type][identifier]
            return

        if callable(PyromodConfig.stopped_handler):
            PyromodConfig.stopped_handler(identifier, listener)
        elif PyromodConfig.throw_exceptions:
            listener["future"].set_exception(ListenerStopped())

        del self.listeners[listener_type][identifier]


@patch(pyrogram.handlers.message_handler.MessageHandler)
class MessageHandler:
    @patchable
    def __init__(self, callback: Callable, filters=None):
        self.registered_handler = callback
        self.old__init__(self.resolve_future, filters)

    @patchable
    async def check(self, client, message):
        if user := getattr(message, "from_user", None):
            user = user.id
        listener = client.match_listener(
            (message.chat.id, user, message.id),
            ListenerTypes.MESSAGE,
        )[0]

        listener_does_match = handler_does_match = False

        if listener:
            filters = listener["filters"]
            if callable(filters):
                if iscoroutinefunction(filters.__call__):
                    listener_does_match = await filters(client, message)
                else:
                    listener_does_match = await client.loop.run_in_executor(
                        None, filters, client, message
                    )
            else:
                listener_does_match = True

        if callable(self.filters):
            if iscoroutinefunction(self.filters.__call__):
                handler_does_match = await self.filters(client, message)
            else:
                handler_does_match = await client.loop.run_in_executor(
                    None, self.filters, client, message
                )
        else:
            handler_does_match = True

        # let handler get the chance to handle if listener
        # exists but its filters doesn't match
        return listener_does_match or handler_does_match

    @patchable
    async def resolve_future(self, client, message, *args):
        listener_type = ListenerTypes.MESSAGE
        if user := getattr(message, "from_user", None):
            user = user.id
        listener, identifier = client.match_listener(
            (message.chat.id, user, message.id),
            listener_type,
        )
        listener_does_match = False
        if listener:
            filters = listener["filters"]
            if callable(filters):
                if iscoroutinefunction(filters.__call__):
                    listener_does_match = await filters(client, message)
                else:
                    listener_does_match = await client.loop.run_in_executor(
                        None, filters, client, message
                    )
            else:
                listener_does_match = True

        if listener_does_match:
            if not listener["future"].done():
                listener["future"].set_result(message)
                del client.listeners[listener_type][identifier]
                raise pyrogram.StopPropagation
        else:
            await self.registered_handler(client, message, *args)


@patch(pyrogram.handlers.callback_query_handler.CallbackQueryHandler)
class CallbackQueryHandler:
    @patchable
    def __init__(self, callback: Callable, filters=None):
        self.registered_handler = callback
        self.old__init__(self.resolve_future, filters)

    @patchable
    async def check(self, client, query):
        chatID, mID = None, None
        if message := getattr(query, "message", None):
            chatID, mID = message.chat.id, message.id
        listener = client.match_listener(
            (chatID, query.from_user.id, mID),
            ListenerTypes.CALLBACK_QUERY,
        )[0]

        # managing unallowed user clicks
        if PyromodConfig.unallowed_click_alert:
            permissive_listener = client.match_listener(
                identifier_pattern=(
                    chatID,
                    None,
                    mID,
                ),
                listener_type=ListenerTypes.CALLBACK_QUERY,
            )[0]

            if (permissive_listener and not listener) and permissive_listener[
                "unallowed_click_alert"
            ]:
                alert = (
                    permissive_listener["unallowed_click_alert"]
                    if type(permissive_listener["unallowed_click_alert"]) == str
                    else PyromodConfig.unallowed_click_alert_text
                )
                try:
                    await query.answer(alert)
                except QueryIdInvalid:
                    return False
                return False

        filters = listener["filters"] if listener else self.filters

        if callable(filters):
            if iscoroutinefunction(filters.__call__):
                return await filters(client, query)
            else:
                return await client.loop.run_in_executor(None, filters, client, query)
        else:
            return True

    @patchable
    async def resolve_future(self, client, query, *args):
        listener_type = ListenerTypes.CALLBACK_QUERY
        chatID, mID = None, None
        if message := getattr(query, "message", None):
            chatID, mID = message.chat.id, message.id
        listener, identifier = client.match_listener(
            (chatID, query.from_user.id, mID),
            listener_type,
        )

        if listener and not listener["future"].done():
            listener["future"].set_result(query)
            del client.listeners[listener_type][identifier]
        else:
            await self.registered_handler(client, query, *args)


@patch(pyrogram.types.messages_and_media.message.Message)
class Message(pyrogram.types.messages_and_media.message.Message):
    @patchable
    async def wait_for_click(
        self,
        from_user_id: Optional[int] = None,
        timeout: Optional[int] = None,
        filters=None,
        alert: Union[str, bool] = True,
    ):
        return await self._client.listen(
            (self.chat.id, from_user_id, self.id),
            listener_type=ListenerTypes.CALLBACK_QUERY,
            timeout=timeout,
            filters=filters,
            unallowed_click_alert=alert,
        )


@patch(pyrogram.types.user_and_chats.chat.Chat)
class Chat(pyrogram.types.Chat):
    @patchable
    def listen(self, *args, **kwargs):
        return self._client.listen((self.id, None, None), *args, **kwargs)

    @patchable
    def ask(self, text, *args, **kwargs):
        return self._client.ask(text, (self.id, None, None), *args, **kwargs)

    @patchable
    def stop_listening(self, *args, **kwargs):
        return self._client.stop_listening(
            *args, identifier_pattern=(self.id, None, None), **kwargs
        )


@patch(pyrogram.types.user_and_chats.user.User)
class User(pyrogram.types.User):
    @patchable
    def listen(self, *args, **kwargs):
        return self._client.listen((None, self.id, None), *args, **kwargs)

    @patchable
    def ask(self, text, *args, **kwargs):
        return self._client.ask(text, (self.id, self.id, None), *args, **kwargs)

    @patchable
    def stop_listening(self, *args, **kwargs):
        return self._client.stop_listening(
            *args, identifier_pattern=(None, self.id, None), **kwargs
        )



from pyrogram.types import (
    ForceReply,
    KeyboardButton,
    ReplyKeyboardMarkup,
    ReplyKeyboardRemove,
)


class ReplyKeyboard(ReplyKeyboardMarkup):
    def __init__(
        self,
        resize_keyboard=None,
        one_time_keyboard=None,
        selective=None,
        placeholder=None,
        row_width=3,
    ):
        self.keyboard = list()
        super().__init__(
            keyboard=self.keyboard,
            resize_keyboard=resize_keyboard,
            one_time_keyboard=one_time_keyboard,
            selective=selective,
            placeholder=placeholder,
        )
        self.row_width = row_width

    def add(self, *args):
        self.keyboard = [
            args[i : i + self.row_width] for i in range(0, len(args), self.row_width)
        ]

    def row(self, *args):
        self.keyboard.append([button for button in args])


class ReplyButton(KeyboardButton):
    def __init__(self, text=None, request_contact=None, request_location=None):
        super().__init__(
            text=text,
            request_contact=request_contact,
            request_location=request_location,
        )


class ReplyKeyboardRemove(ReplyKeyboardRemove):
    def __init__(self, selective=None):
        super().__init__(selective=selective)


class ForceReply(ForceReply):
    def __init__(self, selective=None, placeholder=None):
        super().__init__(selective=selective, placeholder=placeholder)



from functools import partial, wraps
from typing import Iterable, Optional, Union

from pyrogram import Client
from pyrogram.enums import ChatMemberStatus, ChatType
from pyrogram.types import CallbackQuery, Message

group_types: Iterable[ChatType] = (ChatType.GROUP, ChatType.SUPERGROUP)

admin_status: Iterable[ChatMemberStatus] = (
    ChatMemberStatus.OWNER,
    ChatMemberStatus.ADMINISTRATOR,
)


async def check_perms(
    message: Union[CallbackQuery, Message],
    permissions: Optional[Union[list, str]],
    complain_perms: bool,
) -> bool:
    if isinstance(message, CallbackQuery):
        sender = partial(message.answer, show_alert=True)
        chat = message.message.chat
    else:
        sender = message.reply_text
        chat = message.chat

    # TODO : ғᴏʀ ᴀɴᴏɴʏᴍᴏᴜs ᴀᴅᴍɪɴ :( ɪғ ᴜ ᴄᴀɴ ᴛʀʏ ᴘᴜʟʟ

    user = await chat.get_member(message.from_user.id)
    if user.status == ChatMemberStatus.OWNER:
        return True

    # ɴᴏ ᴘᴇʀᴍɪssɪᴏɴs sᴘᴇᴄɪғɪᴇᴅ, ᴀᴄᴄᴇᴘᴛ ʙᴇɪɴɢ ᴀɴ ᴀᴅᴍɪɴ.
    if not permissions and user.status == ChatMemberStatus.ADMINISTRATOR:
        return True

    if user.status != ChatMemberStatus.ADMINISTRATOR:
        if complain_perms:
            await sender("ʏᴏᴜ ᴍᴜsᴛ ʙᴇ ᴀɴ ᴀᴅᴍɪɴɪsᴛʀᴀᴛᴏʀ ᴛᴏ ᴜsᴇ ᴛʜɪs ᴄᴏᴍᴍᴀɴᴅ.")
        return False

    if isinstance(permissions, str):
        permissions = [permissions]

    missing_perms = [
        permission
        for permission in permissions
        if not getattr(user.privileges, permission)
    ]

    if not missing_perms:
        return True
    if complain_perms:
        await sender(
            "ʏᴏᴜ ᴅᴏɴ'ᴛ ʜᴀᴠᴇ ᴛʜᴇ ʀᴇǫᴜɪʀᴇᴅ ᴘᴇʀᴍɪssɪᴏɴs: {permissions}".format(
                permissions=", ".join(missing_perms)
            )
        )
    return False


"""
@adminsOnly("permissions")

can_manage_chat (bool, optional) – True, if the administrator can access the chat event log, chat statistics, message statistics in channels, see channel members, see anonymous administrators in supergroups and ignore slow mode. Implied by any other administrator privilege.

can_delete_messages (bool, optional) – True, if the administrator can delete messages of other users.

can_manage_video_chats (bool, optional) – Groups and supergroups only. True, if the administrator can manage video chats (also called group calls).

can_restrict_members (bool, optional) – True, if the administrator can restrict, ban or unban chat members.

can_promote_members (bool, optional) – True, if the administrator can add new administrators with a subset of his own privileges or demote administrators that he has promoted, directly or indirectly (promoted by administrators that were appointed by the user).

can_change_info (bool, optional) – True, if the user is allowed to change the chat title, photo and other settings.

can_post_messages (bool, optional) – Channels only. True, if the administrator can post messages in the channel.

can_edit_messages (bool, optional) – Channels only. True, if the administrator can edit messages of other users and can pin messages.

can_invite_users (bool, optional) – True, if the user is allowed to invite new users to the chat.

can_pin_messages (bool, optional) – Groups and supergroups only. True, if the user is allowed to pin messages.

is_anonymous (bool, optional) – True, if the user’s presence in the chat is hidden.
"""


def adminsOnly(
    permissions: Union[list, str] = None,
    allow_pm: bool = False,
    complain_perms: bool = True,
):
    def decorator(func):
        @wraps(func)
        async def wrapper(
            client: Client, message: Union[CallbackQuery, Message], *args, **kwargs
        ):
            if isinstance(message, CallbackQuery):
                sender = partial(message.answer, show_alert=True)
                msg = message.message
            elif isinstance(message, Message):
                sender = message.reply_text
                msg = message
            else:
                raise NotImplementedError(
                    f"require_admin ᴄᴀɴ'ᴛ ᴘʀᴏᴄᴇss ᴜᴘᴅᴀᴛᴇs ᴡɪᴛʜ ᴛʜᴇ ᴛʏᴘᴇ '{message.__name__}' yet."
                )

            if msg.chat.type == ChatType.PRIVATE:
                if allow_pm:
                    return await func(client, message, *args, *kwargs)
                return await sender(
                    "ᴛʜɪs ᴄᴏᴍᴍᴀɴᴅ ᴄᴀɴ'ᴛ ʙᴇ ᴜsᴇᴅ ɪɴ ᴀ ᴘʀɪᴠᴀᴛᴇ ᴄʜᴀᴛ. ɪғ ʏᴏᴜ ɴᴇᴇᴅ ᴀɴʏ ʜᴇʟᴘ, ᴘʟᴇᴀsᴇ ᴜsᴇ ᴛʜᴇ <code>/help</code> ᴄᴏᴍᴍᴀɴᴅ"
                )
            if msg.chat.type == ChatType.CHANNEL:
                return await func(client, message, *args, *kwargs)
            has_perms = await check_perms(message, permissions, complain_perms)
            if has_perms:
                return await func(client, message, *args, *kwargs)

        return wrapper

    return decorator
